def quote(value: str) -> str:
    """
    Enclose a value in quotes

    :param value: Value to enclosed
    :return: Quoted value

    """

    if '"' in value:
        if "'" in value:
            quote_chars = '"""'
            if value.startswith('"'):
                value = f"\\{value}"
        else:
            quote_chars = "'"
    else:
        quote_chars = '"'

    return f"{quote_chars}{value}{quote_chars}"



# pylint hates pydantic
# pylint: disable=E0213,E0611
from typing import List, Dict, Tuple, Union, Any, Callable
from functools import partial

from pydantic import BaseModel, validator, constr, conlist

from glotter.utils import quote

TransformationFuncT = Callable[[str, str], Tuple[str, str]]


class AutoGenParam(BaseModel):
    """Object used for an auto-generated test parameter"""

    name: constr(min_length=1)
    input: Union[None, str]
    expected: Union[str, List[str], Dict[str, str]]


class AutoGenTest(BaseModel):
    """Object containing information about an auto-generated test"""

    name: constr(min_length=1, regex="^[a-zA-Z][0-9a-zA-Z_]*$")
    requires_parameters: bool = False
    params: conlist(AutoGenParam, min_items=1)
    transformations: List[TransformationFuncT] = []

    @validator("params", each_item=True, pre=True)
    def validate_params(
        cls, value: Dict[str, Any], values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Validate each parameter

        :param value: Parameter to validate
        :param values: Test item
        :return: Original parameter if project requires parameters. Otherwise, parameter with
            an empty "name" and an "input" of `None`
        :raises: :exc:`ValueError` if project requires parameters but no input
        """

        if values.get("requires_parameters"):
            if "input" not in value:
                raise ValueError(
                    'This project requires parameters, but "input" is not specified'
                )

            return value

        return {**value, "name": value.get("name") or "na", "input": value.get("input")}

    @validator("transformations", each_item=True, pre=True)
    def validate_transformation(cls, value: Any) -> TransformationFuncT:
        """
        Validate each transformation

        :param value: Transformation to validate
        :return: Transformation function
        :raises: :exc:`ValueError` if invalid transformation
        """

        scalar_transformation_funcs = {
            "strip": partial(_append_method_to_actual, "strip"),
            "splitlines": partial(_append_method_to_actual, "splitlines"),
            "lower": partial(_append_method_to_actual, "lower"),
            "any_order": _unique_sort,
            "strip_expected": partial(_append_method_to_expected, "strip"),
        }
        dict_transformation_funcs = {
            "remove": {
                "validator": IsListOfStr,
                "func": _remove_chars,
            },
            "strip": {
                "validator": IsListOfStr,
                "func": _strip_chars,
            },
        }

        try:
            key = value
            if isinstance(value, str):
                return scalar_transformation_funcs[value]
            elif isinstance(value, dict):
                key, item = tuple(*value.items())
                dict_transformation_funcs[key]["validator"](item=item)
                return partial(dict_transformation_funcs[key]["func"], item)
        except KeyError as e:
            raise ValueError(f'Invalid transformation "{key}"') from e

        raise ValueError(f'Invalid transformation data type "{type(value).__name__}"')


class IsListOfStr(BaseModel):
    """Data model to make sure that the transformation value is a list of strings"""

    item: List[str]


def _append_method_to_actual(
    method: str, actual_var: str, expected_var
) -> Tuple[str, str]:
    return f"{actual_var}.{method}()", expected_var


def _append_method_to_expected(
    method: str, actual_var: str, expected_var: str
) -> Tuple[str, str]:
    return actual_var, f"{expected_var}.{method}()"


def _remove_chars(
    values: List[str], actual_var: str, expected_var: str
) -> Tuple[str, str]:
    for value in values:
        actual_var += f'.replace({quote(value)}, "")'

    return actual_var, expected_var


def _strip_chars(
    values: List[str], actual_var: str, expected_var: str
) -> Tuple[str, str]:
    for value in values:
        actual_var += f".strip({quote(value)})"

    return actual_var, expected_var


def _unique_sort(actual_var, expected_var):
    return f"sorted(set({actual_var}))", f"sorted(set({expected_var}))"



import os
import shutil

from black import format_str, Mode

from glotter.settings import Settings
from glotter.utils import quote

AUTO_GEN_TEST_PATH = "test/generated"


def generate_tests():
    """
    Generate tests for all projects
    """

    shutil.rmtree(AUTO_GEN_TEST_PATH, ignore_errors=True)
    settings = Settings()
    test_generators = {
        project_name: TestGenerator(project_name, project)
        for project_name, project in settings.projects.items()
    }
    test_codes = {}
    for project_name, test_generator in test_generators.items():
        test_code = test_generator.generate_tests()
        if test_code:
            test_codes[project_name] = test_code

    for project_name, test_code in test_codes.items():
        test_generators[project_name].write_tests(test_code)


class TestGenerator:
    __test__ = False  # Indicate this is not a test

    def __init__(self, project_name, project):
        self.project_name = project_name
        self.project = project
        self.long_project_name = "_".join(self.project.words)

    def generate_tests(self):
        if not self.project.tests:
            return ""

        test_code = self._get_imports() + self._get_project_fixture()
        for test_obj in self.project.tests:
            test_code += self._generate_test(test_obj)

        return format_str(test_code, mode=Mode())

    def _get_imports(self):
        test_code = "from glotter import project_test, project_fixture\n"
        if self.project.requires_parameters:
            test_code += "import pytest\n"

        return test_code

    def _get_project_fixture(self):
        return f"""\
PROJECT_NAME="{self.project_name}"
@project_fixture(PROJECT_NAME)
def {self.long_project_name}(request):
    request.param.build()
    yield request.param
    request.param.cleanup()
"""

    def _generate_test(self, test_obj):
        test_code = "@project_test(PROJECT_NAME)\n"
        func_params = ""
        run_param = ""
        if self.project.requires_parameters:
            test_code += self._generate_params(test_obj)
            func_params = "in_params, expected, "
            run_param = "params=in_params"

        test_code += self._get_test_function_and_run(test_obj, func_params, run_param)
        test_code += _indent(self._get_expected_output(test_obj), 4)
        actual_var, expected_var = self._get_transformation_vars(test_obj)
        test_code += _indent(
            _get_assert(actual_var, expected_var, test_obj.params[0].expected), 4
        )
        return test_code

    def _generate_params(self, test_obj):
        pytest_params = "".join(
            _indent(self._generate_param(param), 8) for param in test_obj.params
        ).strip()
        return f"""\
@pytest.mark.parametrize(
    ("in_params", "expected"),
    [
        {pytest_params}
    ]
)
"""

    def _generate_param(self, param):
        input_param = param.input
        if isinstance(input_param, str):
            input_param = quote(input_param)

        expected_output = param.expected
        if isinstance(expected_output, str):
            expected_output = quote(expected_output)

        return f'pytest.param({input_param}, {expected_output}, id="{param.name}"),\n'

    def _get_test_function_and_run(self, test_obj, func_params, run_param):
        return f"""\
def test_{test_obj.name}({func_params}{self.long_project_name}):
    actual = {self.long_project_name}.run({run_param})
"""

    def _get_expected_output(self, test_obj):
        if self.project.requires_parameters:
            return ""

        expected_output = test_obj.params[0].expected
        if isinstance(expected_output, dict):
            return self._generate_expected_file(expected_output)
        elif isinstance(expected_output, str):
            expected_output = quote(expected_output)
        else:
            expected_output = str(expected_output)

        return f"expected = {expected_output}\n"

    def _generate_expected_file(self, expected_output):
        test_code = ""
        if "exec" in expected_output:
            script = quote(expected_output["exec"])
            test_code = f"expected = {self.long_project_name}.exec({script})\n"
        elif "self" in expected_output:
            test_code = f"""\
with open({self.long_project_name}.full_path, "r", encoding="utf-8") as file:
    expected = file.read()
"""
        return test_code

    def _get_transformation_vars(self, test_obj):
        actual_var = "actual"
        expected_var = "expected"
        for transformation in test_obj.transformations:
            actual_var, expected_var = transformation(actual_var, expected_var)

        return actual_var, expected_var

    def write_tests(self, test_code):
        os.makedirs(AUTO_GEN_TEST_PATH, exist_ok=True)
        with open(
            os.path.join(AUTO_GEN_TEST_PATH, f"test_{self.long_project_name}.py"),
            "w",
            encoding="utf-8",
        ) as f:
            f.write(test_code)


def _indent(str_value, num_spaces):
    spaces = " " * num_spaces
    return "".join(f"{spaces}{line}" for line in str_value.splitlines(keepends=True))


def _get_assert(actual_var, expected_var, expected_output):
    if isinstance(expected_output, list):
        return f"""\
actual_list = {actual_var}
expected_list = {expected_var}
assert len(actual_list) == len(expected_list), "Length not equal"
for index in range(len(expected_list)):
    assert actual_list[index] == expected_list[index], f"Item {{index + 1}} is not equal"
"""

    test_code = ""
    if actual_var != "actual":
        test_code += f"actual = {actual_var}\n"

    if expected_var != "expected":
        test_code += f"expected = {expected_var}\n"

    return f"{test_code}assert actual == expected\n"
